# Copyright (c) 2024 Rouast Labs
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import numpy as np
import os
import sys
from typing import Union
import urllib.request
import yaml

if sys.version_info >= (3, 9):
  from importlib.resources import open_binary
else:
  from importlib_resources import open_binary

from vitallens.constants import API_MIN_FRAMES, VIDEO_PARSE_ERROR

def load_config(filename: str) -> dict:
  """Load a yaml config file.

  Args:
    filename: The filename of the yaml config file
  Returns:
    loaded: The contents of the yaml config file
  """
  with open_binary('vitallens.configs', filename) as f:
    loaded = yaml.load(f, Loader=yaml.Loader)
  return loaded

def download_file(url: str, dest: str):
  """Download a file if necessary.

  Args:
    url: The url to download the file from
    dest: The path to write the downloaded file to
  """
  if not os.path.exists(dest):
    os.makedirs(os.path.dirname(dest), exist_ok=True)
    logging.info(f"Downloading {url} to {dest}")
    urllib.request.urlretrieve(url, dest)
  else:
    logging.info(f"{dest} already exists, skipping download.")

def merge_faces(faces: np.ndarray) -> tuple:
  """Compute the union of all faces.
  
  Args:
    faces: The face detections. Shape (n, 4)
  Returns:
    union: Tuple (x0, y0, x1, y1)
  """
  # Get the minimum x0, y0 and maximum x1, y1 values across all frames
  x0 = np.min(faces[:, 0])
  y0 = np.min(faces[:, 1])
  x1 = np.max(faces[:, 2])
  y1 = np.max(faces[:, 3])
  # Return as tuple
  return (x0, y0, x1, y1)

def check_faces(
    faces: Union[list, np.ndarray],
    inputs_shape: tuple
  ) -> np.ndarray:
  """Make sure the face detections are in a correct format.

  Args:
    faces: The specified faces in form [x0, y0, x1, y1]. Either
      - list/ndarray of shape (n_faces, n_frames, 4) for multiple faces detected on multiple frames
      - list/ndarray of shape (n_frames, 4) for single face detected on mulitple frames
      - list/ndarray of shape (4,) for single face detected globally
      - None to assume frames already cropped to single face
    inputs_shape: The shape of the inputs.
  Returns:
    faces: The faces. np.ndarray of shape (n_faces, n_frames, 4) in form [x_0, y_0, x_1, y_1]
  """
  n_frames, h, w, _ = inputs_shape
  if faces is None:
    # Assume that each entire frame is a single face
    logging.info("No faces given - assuming that frames have been cropped to a single face")
    faces = np.tile(np.asarray([0, 0, w, h], dtype=np.int64), (n_frames, 1))[np.newaxis] # (1, n_frames, 4)
  else:
    faces = np.asarray(faces, dtype=np.int64)
    if faces.shape[-1] != 4: raise ValueError("Face detections must be in flat point form")
    if len(faces.shape) == 1:
      # Single face detection given - repeat for n_frames
      faces = np.tile(faces, (n_frames, 1))[np.newaxis] # (1, n_frames, 4)
    elif len(faces.shape) == 2:
      # Single face detections for multiple frames given
      if faces.shape[0] != n_frames:
        raise ValueError(f"Assuming detections of a single face for multiple frames given, but number of frames ({n_frames}) did not match number of face detections ({faces.shape[0]})")
      faces = faces[np.newaxis]
    elif len(faces.shape) == 3:
      if faces.shape[1] == 1:
        # Multiple face detections for single frame given
        faces = np.tile(faces, (1, n_frames, 1)) # (n_faces, n_frames, 4)
      else:
        # Multiple face detections for multiple frames given
        if faces.shape[1] != n_frames:
          raise ValueError(f"Assuming detections of multiple faces for multiple frames given, but number of frames ({n_frames}) did not match number of detections for each face ({faces.shape[1]})")
  # Check that x0 < x1 and y0 < y1 for all faces
  if not (np.all((faces[...,2] - faces[...,0]) > 0) and np.all((faces[...,3] - faces[...,1]) > 0)):
    raise ValueError("Face detections are invalid, should be in form [x0, y0, x1, y1]")
  return faces

def check_faces_in_roi(
    faces: np.ndarray,
    roi: Union[np.ndarray, tuple, list],
    percentage_required_inside_roi: tuple = (0.5, 0.5)
  ) -> bool:
  """Check whether all faces are sufficiently inside the ROI.

  Args:
    faces: The faces. Shape (n_faces, 4) in form (x0, y0, x1, y1)
    roi: The region of interest. Shape (4,) in form (x0, y0, x1, y1)
    percentage_required_inside_roi: Tuple (w, h) indicating what percentage
      of width/height of face is required to remain inside the ROI.
  Returns:
    out: True if all faces are sufficiently inside the ROI.
  """
  faces_w = faces[:,2] - faces[:,0]
  faces_h = faces[:,3] - faces[:,1]
  faces_inside_roi = np.logical_and(
    np.logical_and(faces[:,2] - roi[0] > percentage_required_inside_roi[0] * faces_w,
                   roi[2] - faces[:,0] > percentage_required_inside_roi[0] * faces_w),
    np.logical_and(faces[:,3] - roi[1] > percentage_required_inside_roi[1] * faces_h,
                   roi[3] - faces[:,1] > percentage_required_inside_roi[1] * faces_h))
  facess_inside_roi = np.all(faces_inside_roi)
  return facess_inside_roi

def convert_ndarray_to_list(d: Union[dict, list, np.ndarray]):
  """Recursively convert any np.ndarray to list in nested object.
  
  Args:
    d: Nested object consisting of list, dict, and np.ndarray
  Returns:
    out: The same object with any np.ndarray converted to list
  """
  if isinstance(d, np.ndarray):
    return d.tolist()
  elif isinstance(d, dict):
    return {k: convert_ndarray_to_list(v) for k, v in d.items()}
  elif isinstance(d, list):
    return [convert_ndarray_to_list(i) for i in d]
  else:
    return d
  


# Copyright (c) 2024 Rouast Labs
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import itertools
import logging
import math
import numpy as np
import os
from prpy.numpy.image import parse_image_inputs
from prpy.numpy.interp import interpolate_vals
import sys
from typing import Tuple

if sys.version_info >= (3, 9):
  from importlib.resources import files
else:
  from importlib_resources import files

INPUT_SIZE = (240, 320)
MAX_SCAN_FRAMES = 60

def nms(
    boxes: np.ndarray,
    scores: np.ndarray,
    max_output_size: int,
    iou_threshold: float,
    score_threshold: float
  ) -> Tuple[np.ndarray, np.ndarray]:
  """Non-maximum suppression for a batch of box predictions.

  Args:
    boxes: Boxes in point form [0, 1]. Shape (n_batch, n_anchors, 4)
    scores: Probability scores. Shape (n_batch, n_anchors)
    max_output_size: Maximum number of boxes to be selected. Scalar.
    iou_threshold: Threshold wrt iou for amount of box overlap. Scalar.
    score_threshold: Threshold wrt score for removing boxes. Scalar.
  Returns:
    Tuple of
     - idxs: The selected indices padded with zero. Shape (n_batch, max_output_size)
     - num_valid: Number of valid elements per batch. Shape (n_batch,)
  """
  n_batch = boxes.shape[0]
  # Split up box coordinates
  x1 = boxes[..., 0]
  y1 = boxes[..., 1]
  x2 = boxes[..., 2]
  y2 = boxes[..., 3]
  # Compute the areas for all boxes
  areas = (x2 - x1) * (y2 - y1)
  # Initialise result
  out_idxs = np.zeros((n_batch, max_output_size), dtype=np.int64)
  num_valid = []
  # Do nms separately for each batch element
  for b in range(n_batch):
    # Get sorted idxs for scores high->low
    idxs = scores[b].argsort()[::-1]
    # Iterate through idxs
    result_b = []
    while idxs.size > 0:
      i = idxs[0]
      # Stop when best score falls under threshold
      if scores[b,i] < score_threshold or len(result_b) == max_output_size:
        break
      # Keep best scoring idx
      result_b.append(i)
      # Calculate IOU between this idx and all others
      xx1 = np.maximum(x1[b,i], x1[b,idxs[1:]])
      yy1 = np.maximum(y1[b,i], y1[b,idxs[1:]])
      xx2 = np.minimum(x2[b,i], x2[b,idxs[1:]])
      yy2 = np.minimum(y2[b,i], y2[b,idxs[1:]])
      w = np.maximum(0.0, xx2 - xx1)
      h = np.maximum(0.0, yy2 - yy1)
      inters = w * h
      ious = inters / (areas[b,i] + areas[b,idxs[1:]] - inters)
      # Keep only idxs for which IOU is under threshold
      keep = np.where(ious <= iou_threshold)[0]
      idxs = idxs[keep + 1]
    # Write result_b to result
    out_idxs[b,0:len(result_b)] = result_b
    num_valid.append(len(result_b))
  return out_idxs, np.asarray(num_valid)

def enforce_temporal_consistency(
    boxes: np.ndarray,
    info: np.ndarray,
    n_frames: int
  ) -> Tuple[np.ndarray, np.ndarray]:
  """Enforce temporal consistency by sorting faces along second axis to minimize spatial distances.
  
  Args:
    boxes: Detected boxes in point form [0, 1], shape (n_frames, n_faces, 4)
    info: Detection info: idx, scanned, scan_found_face, confidence. Shape (n_frames, n_faces, 4)
    n_frames: Number of frames in the original input.
  Returns:
    Tuple of
     - boxes: Processed boxes in point form [0, 1], shape (n_frames, n_faces, 4)
     - info: Processed info: idx, scanned, scan_found_face, confidence. Shape (n_frames, n_faces, 4)
  """
  # Make sure that enough frames are present
  if n_frames == 1:
    return boxes, info
  # Determine the maximum number of detections in any frame
  max_det_faces = int(np.max(np.sum(info[...,2], axis=-1)))
  # Trim to max_det_faces
  boxes = boxes[:,0:max_det_faces]
  info = info[:,0:max_det_faces]
  def distance_minimizing_idxs(boxes, info, max_det_faces):
    """Compute index permutations along second axis that minimize distances"""
    centers = (boxes[...,2:] + boxes[...,:2])/2
    valid = info[...,2]
    idx_perms = list(itertools.permutations(range(max_det_faces)))
    prev_c = centers[0]
    prev_idx_perm = idx_perms[0]
    idxs = []
    for c, v in zip(centers, valid):
      # Assume last known position for any non-valid faces.
      c[v == 0] = prev_c[v == 0]
      # Order according to prev_idx_perm
      c_ = np.take(c, prev_idx_perm, axis=0)
      # Compute distances at all possible permutations
      dists = [np.mean(np.linalg.norm(prev_c - np.take(c_, p, axis=0), axis=1)) for p in idx_perms]
      # Select permutation of idxs that minimizes distance to prev_centers
      idx_perm = idx_perms[np.argmin(dists)]
      idxs.append(idx_perm)
      prev_c, prev_idx_perm = c, idx_perm
    return np.asarray(idxs)[...,np.newaxis]
  # Make second axis temporally consistent by minimizing distances between centers
  idxs = distance_minimizing_idxs(boxes, info, max_det_faces)
  boxes = np.take_along_axis(boxes, idxs, axis=1)
  info = np.take_along_axis(info, idxs, axis=1)
  # Sort second axis by total confidence
  order = np.argsort(np.sum(info[...,3], axis=0))[::-1][np.newaxis]
  boxes = np.squeeze(np.take(boxes, order, axis=1), axis=1)
  info = np.squeeze(np.take(info, order, axis=1), axis=1)
  return boxes, info

def interpolate_unscanned_frames(
    boxes: np.ndarray,
    info: np.ndarray,
    n_frames: int
  ) -> Tuple[np.ndarray, np.ndarray]:
  """Interpolate values for frames that were not scanned.

  Args:
    boxes: Detected boxes in point form [0, 1], shape (n_frames, n_faces, 4)
    info: Detection info: idx, scanned, scan_found_face, interp_valid, confidence. Shape (n_frames, n_faces, 5)
    n_frames: Number of frames in the original input.
  Returns:
    Tuple of
     - boxes: Processed boxes in point form [0, 1], shape (orig_n_frames, n_faces, 4)
     - info: Processed info: idx, scanned, scan_found_face, confidence. Shape (orig_n_frames, n_faces, 4)
  """
  _, n_faces, _ = info.shape
  # Add rows corresponding to unscanned frames
  add_idxs = list(set.difference(set(range(n_frames)), set(info[:,0,0].astype(np.int32).tolist())))
  idxs = np.repeat(np.asarray(add_idxs)[:,np.newaxis], n_faces, axis=1)[...,np.newaxis]
  # Info
  add_info = np.r_['2', idxs, np.zeros_like(idxs, np.int32), np.zeros_like(idxs, np.int32), np.zeros_like(idxs, np.int32)]
  info = np.concatenate([info, add_info])
  # Boxes
  add_boxes = np.full([len(add_idxs), n_faces, 4], np.nan)
  boxes = np.concatenate([boxes, add_boxes])
  # Sort according to frame idxs
  sort_idxs = np.argsort(info[:,0,0])
  boxes = np.take(boxes, sort_idxs, axis=0)
  info = np.take(info, sort_idxs, axis=0)
  # Interpolation
  boxes = np.apply_along_axis(interpolate_vals, 0, boxes)
  return boxes, info

class FaceDetector:
  def __init__(
      self,
      max_faces: int,
      fs: float,
      score_threshold: float,
      iou_threshold: float):
    """Initialise the face detector.

    Args:
      max_faces: The maximum number of faces to detect.
      fs: Frequency [Hz] at which faces should be detected. Detections are
        linearly interpolated for remaining frames.
      score_threshold: Face detection score threshold.
      iou_threshold: Face detection iou threshold.
    """
    import onnxruntime as rt
    with files('vitallens.models.Ultra-Light-Fast-Generic-Face-Detector-1MB') as model_dir:
      self.model = rt.InferenceSession(os.path.join(model_dir, "model_rfb_320.onnx"), providers=['CPUExecutionProvider'])
    self.iou_threshold = iou_threshold if iou_threshold is not None else self.config['threshold']
    self.score_threshold = score_threshold
    self.max_faces = max_faces
    self.fs = fs
  def __call__(
      self,
      inputs: Tuple[np.ndarray, str],
      n_frames: int,
      fps: float
    ) -> Tuple[np.ndarray, np.ndarray]:
    """Run inference.

    Args:
      inputs: The video to analyze. Either a np.ndarray of shape (n_frames, h, w, 3)
        with a sequence of frames in unscaled uint8 RGB format, or a path to a video file.
      n_frames: The number of input video frames
      fps: Sampling frequency of the input video.
    Returns:
      Tuple of
       - boxes: Detected face boxes in relative flat point form (n_frames, n_faces, 4)
       - info: Tuple (idx, scanned, scan_found_face, interp_valid, confidence) (n_frames, n_faces, 5)
    """
    # Determine number of batches
    n_batches = math.ceil((n_frames / (fps / self.fs)) / MAX_SCAN_FRAMES)
    if n_batches > 1:
      logging.info(f"Running face detection in {n_batches} batches...")
    # Determine frame offsets for batches
    offsets_lengths = [(i[0], len(i)) for i in np.array_split(np.arange(n_frames), n_batches)]
    # Process in batches
    results = [self.scan_batch(inputs=inputs, batch=i, n_batches=n_batches, start=int(s), end=int(s+l), fps=fps) for i, (s, l) in enumerate(offsets_lengths)]
    if len(results) == 1:
      boxes, classes, scan_idxs = results[0]
      scan_idxs = np.asarray(scan_idxs)
    else:
      boxes = np.concatenate([r[0] for r in results], axis=0)
      classes = np.concatenate([r[1] for r in results], axis=0)
      scan_idxs = np.concatenate([r[2] for r in results], axis=0)
    scan_every = 1 if len(scan_idxs) == 1 else int(np.max(np.diff(scan_idxs)))
    n_frames_scan = boxes.shape[0]
    # Non-max suppression
    idxs, num_valid = nms(boxes=boxes,
                          scores=classes[..., 1],
                          max_output_size=self.max_faces,
                          iou_threshold=self.iou_threshold,
                          score_threshold=self.score_threshold)
    max_valid = np.max(num_valid)
    idxs = idxs[...,0:max_valid]
    boxes = boxes[np.arange(boxes.shape[0])[:, None], idxs]
    classes = classes[np.arange(classes.shape[0])[:, None], idxs]
    # Check if any faces found
    if max_valid == 0:
      logging.warning("No faces found")
      return [], []
    # Assort info: idx, scanned, scan_found_face, confidence
    idxs = np.repeat(scan_idxs[...,np.newaxis], max_valid, axis=1)[...,np.newaxis]
    scanned = np.ones((n_frames_scan, max_valid, 1), dtype=np.int32)
    scan_found_face = np.where(classes[...,1:2] < self.score_threshold, np.zeros([n_frames_scan, max_valid, 1], dtype=np.int32), scanned)
    info = np.r_['2', idxs, scanned, scan_found_face, classes[...,1:2]]
    # Enforce temporal consistency
    boxes, info = enforce_temporal_consistency(boxes=boxes, info=info, n_frames=n_frames)
    # Interpolate unscanned frames if necessary
    if scan_every > 1:
      # Set unsuccessful detections to nan
      nan = info[:,:,2] == 0
      boxes[nan] = np.nan
      # Interpolate
      boxes, info = interpolate_unscanned_frames(boxes=boxes, info=info, n_frames=n_frames)
    # Return
    return boxes, info
  def scan_batch(
      self,
      batch: int,
      n_batches: int,
      inputs: Tuple[np.ndarray, str],
      start: int,
      end: int,
      fps: float = None,
    ) -> Tuple[np.ndarray, np.ndarray, list]:
    """Parse video and run inference for one batch.

    Args:
      batch: The number of this batch.
      n_batches: The total number of batches.
      inputs: The video to analyze. Either a np.ndarray of shape (n_frames, h, w, 3)
        with a sequence of frames in unscaled uint8 RGB format, or a path to a video file.
      start: The index of first frame of the video to analyze in this batch.
      end: The index of the last frame of the video to analyze in this batch.
      fps: Sampling frequency of the input video. Required if type(video) == np.ndarray.
    Returns:
      Tuple of
       - boxes: Scanned boxes in flat point form (n_frames, n_boxes, 4)
       - classes: Detection scores for boxes (n_frames, n_boxes, 2)
       - idxs: Indices of the scanned frames from the original video
    """
    logging.debug(f"Batch {batch}/{n_batches}...")
    # Parse the inputs
    inputs, fps, _, _, idxs = parse_image_inputs(
      inputs=inputs, fps=fps, roi=None, target_size=INPUT_SIZE, target_fps=self.fs,
      preserve_aspect_ratio=False, library='prpy', scale_algorithm='bilinear', 
      trim=(start, end), allow_image=False, videodims=True)
    # Forward pass
    onnx_inputs = {"args_0": (inputs.astype(np.float32) - 127.0) / 128.0}
    onnx_outputs = self.model.run(None, onnx_inputs)[0]
    boxes = onnx_outputs[..., -4:]
    classes = onnx_outputs[..., 0:2]
    # Return
    return boxes, classes, idxs
    


# Copyright (c) 2024 Philipp Rouast
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
from prpy.numpy.image import parse_image_inputs, probe_image_inputs
import pytest

import sys
sys.path.append('../vitallens-python')

from vitallens.client import Method
from vitallens.utils import load_config
from vitallens.utils import merge_faces, check_faces, check_faces_in_roi

@pytest.mark.parametrize("method", [m for m in Method])
def test_load_config(method):
  config = load_config(method.name.lower() + ".yaml")
  assert config['model'] == method.name.lower()

@pytest.mark.parametrize("file", [True, False])
def test_probe_video_inputs(request, file):
  if file:
    test_video_path = request.getfixturevalue('test_video_path')
    video_shape, fps, i = probe_image_inputs(test_video_path)
  else:
    test_video_ndarray = request.getfixturevalue('test_video_ndarray')
    test_video_fps = request.getfixturevalue('test_video_fps')
    video_shape, fps, i = probe_image_inputs(test_video_ndarray, fps=test_video_fps)
  assert video_shape == (360, 480, 768, 3)
  assert fps == 30
  assert i == False

def test_probe_video_inputs_no_file():
  with pytest.raises(Exception):
    _ = probe_image_inputs("does_not_exist", fps="fps")

def test_probe_video_inputs_wrong_fps(request):
  with pytest.raises(Exception):
    test_video_path = request.getfixturevalue('test_video_path')
    _ = probe_image_inputs(test_video_path, fps="fps")

def test_probe_video_inputs_no_fps(request):
  test_video_ndarray = request.getfixturevalue('test_video_ndarray')
  with pytest.raises(Exception):
    _ = probe_image_inputs(test_video_ndarray)

def test_probe_video_inputs_wrong_dtype(request):
  test_video_ndarray = request.getfixturevalue('test_video_ndarray')
  with pytest.raises(Exception):
    _ = probe_image_inputs(test_video_ndarray.astype(np.float32), fps=30.)

def test_probe_video_inputs_wrong_shape_1(request):
  test_video_ndarray = request.getfixturevalue('test_video_ndarray')
  with pytest.raises(Exception):
    _ = probe_image_inputs(test_video_ndarray[np.newaxis], fps=30.)

def test_probe_video_inputs_wrong_shape_2(request):
  test_video_ndarray = request.getfixturevalue('test_video_ndarray')
  with pytest.raises(Exception):
    _ = probe_image_inputs(test_video_ndarray[...,0:1], fps=30.)

def test_probe_video_inputs_wrong_type():
  with pytest.raises(Exception):
    _ = probe_image_inputs(12345, fps=30.)

@pytest.mark.parametrize("file", [True, False])
@pytest.mark.parametrize("roi", [None, (200, 0, 500, 350)])
@pytest.mark.parametrize("target_size", [None, 200])
@pytest.mark.parametrize("target_fps", [None, 15])
def test_parse_video_inputs(request, file, roi, target_size, target_fps):
  if file:
    test_video_path = request.getfixturevalue('test_video_path')
    parsed, fps_in, video_shape_in, ds_factor, idxs = parse_image_inputs(
      inputs=test_video_path, roi=roi, target_size=target_size, target_fps=target_fps)
  else:
    test_video_ndarray = request.getfixturevalue('test_video_ndarray')
    test_video_fps = request.getfixturevalue('test_video_fps')
    parsed, fps_in, video_shape_in, ds_factor, idxs = parse_image_inputs(
      inputs=test_video_ndarray, fps=test_video_fps, roi=roi, target_size=target_size,
      target_fps=target_fps)
  assert parsed.shape == (360 if target_fps is None else 360 // 2,
                          200 if target_size is not None else (350 if roi is not None else 480),
                          200 if target_size is not None else (300 if roi is not None else 768),
                          3)
  assert fps_in == 30
  assert video_shape_in == (360, 480, 768, 3)
  assert ds_factor == 1 if target_fps is None else 2
  assert idxs == list(range(360)) if target_fps is None else list(range(0, 360, 2))

def test_parse_video_inputs_no_file():
  with pytest.raises(Exception):
    _ = parse_image_inputs("does_not_exist")

def test_parse_video_inputs_wrong_type():
  with pytest.raises(Exception):
    _ = parse_image_inputs(12345, fps=30.)

def test_merge_faces():
  np.testing.assert_equal(
    merge_faces(np.asarray([[2, 4, 3, 7],
                            [1, 3, 4, 6],
                            [2, 4, 5, 7],
                            [3, 5, 4, 8]])),
    np.asarray([1, 3, 5, 8]))
  
@pytest.mark.parametrize("faces", [None, [[[3, 5, 4, 8], [3, 5, 4, 8]], [[3, 5, 4, 8], [2, 6, 4, 8]]], [[3, 5, 4, 8], [2, 6, 4, 8]], [3, 5, 4, 8]])
@pytest.mark.parametrize("faces_list", [True, False])
def test_check_faces(faces, faces_list):
  if faces is None and not faces_list:
    pytest.skip("Skip because parameter combination does not work")
  if faces_list:
    faces = check_faces(faces, inputs_shape=(2, 10, 10, 3))
  else:
    faces = check_faces(np.asarray(faces), inputs_shape=(2, 10, 10, 3))
  assert isinstance(faces, np.ndarray)
  assert faces.ndim == 3
  assert faces.shape[1] == 2
  assert faces.shape[2] == 4

def test_check_faces_not_flat_point_form():
  with pytest.raises(Exception):
    _ = check_faces([0, 0, 0, 0, 0], inputs_shape=(2, 10, 10, 3))

def test_check_faces_n_frames_n_dets_not_matching_1():
  with pytest.raises(Exception):
    _ = check_faces([[3, 5, 4, 8]], inputs_shape=(2, 10, 10, 3))

def test_check_faces_n_frames_n_dets_not_matching_2():
  with pytest.raises(Exception):
    _ = check_faces([[[3, 5, 4, 8], [3, 5, 4, 8], [3, 5, 4, 8]]], inputs_shape=(2, 10, 10, 3))

def test_check_faces_invalid_dets():
  with pytest.raises(Exception):
    _ = check_faces([1, 3, 2, 2], inputs_shape=(2, 10, 10, 3))

@pytest.mark.parametrize("scenario", [([[0.4, 0.4, 0.6, 0.6], [0.4, 0.4, 0.6, 0.6]], [0.3, 0.3, 0.7, 0.7], (0.5, 0.5), True), # Fully in ROI
                                      ([[0.2, 0.4, 0.6, 0.6], [0.4, 0.4, 0.6, 0.6]], [0.3, 0.3, 0.7, 0.7], (0.5, 0.5), True), # Slightly out to left but ok
                                      ([[0.0, 0.4, 0.4, 0.6], [0.4, 0.4, 0.6, 0.6]], [0.3, 0.3, 0.7, 0.7], (0.5, 0.5), False), # Too much out to left
                                      ([[0.5, 0.4, 0.9, 0.6], [0.4, 0.4, 0.6, 0.6]], [0.3, 0.3, 0.7, 0.7], (0.2, 0.5), True), # Slightly out to right but ok
                                      ([[0.5, 0.4, 0.9, 0.6], [0.4, 0.4, 0.6, 0.6]], [0.3, 0.3, 0.7, 0.7], (0.6, 0.5), False), # Too much out to right
                                      ([[0.4, 0.2, 0.6, 0.5], [0.4, 0.4, 0.6, 0.6]], [0.3, 0.3, 0.7, 0.7], (0.5, 0.5), True), # Slightly out to top but ok
                                      ([[0.4, 0.1, 0.6, 0.5], [0.4, 0.4, 0.6, 0.6]], [0.3, 0.3, 0.7, 0.7], (0.5, 0.6), False), # Too much out to top
                                      ([[0.4, 0.5, 0.6, 0.8], [0.4, 0.4, 0.6, 0.6]], [0.3, 0.3, 0.7, 0.7], (0.5, 0.5), True), # Slightly out to bottom but ok
                                      ([[0.4, 0.6, 0.6, 0.9], [0.4, 0.4, 0.6, 0.6]], [0.3, 0.3, 0.7, 0.7], (0.5, 0.5), False), # Too much out to bottom
                                      ])
def test_check_faces_in_roi(scenario):
  faces = np.asarray(scenario[0])
  roi = np.asarray(scenario[1])
  percentage_required_inside_roi = scenario[2]
  expected = scenario[3]
  actual = check_faces_in_roi(faces=faces, roi=roi, percentage_required_inside_roi=percentage_required_inside_roi)
  assert expected == actual



# Copyright (c) 2024 Rouast Labs
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from datetime import datetime
import json
import logging
import numpy as np
import os
from prpy.numpy.image import probe_image_inputs
from prpy.numpy.physio import EScope, EMethod
from prpy.numpy.physio import estimate_hr_from_signal, estimate_rr_from_signal
from prpy.numpy.rolling import rolling_calc
from typing import Union

from vitallens.constants import DISCLAIMER, API_MAX_FRAMES
from vitallens.constants import CALC_HR_WINDOW_SIZE, CALC_RR_WINDOW_SIZE
from vitallens.enums import Method, Mode
from vitallens.methods.g import GRPPGMethod
from vitallens.methods.chrom import CHROMRPPGMethod
from vitallens.methods.pos import POSRPPGMethod
from vitallens.methods.vitallens import VitalLensRPPGMethod
from vitallens.ssd import FaceDetector
from vitallens.utils import load_config, check_faces, convert_ndarray_to_list

logging.getLogger().setLevel("INFO")

class VitalLens:
  def __init__(
      self, 
      method: Method = Method.VITALLENS,
      mode: Mode = Mode.BATCH,
      api_key: str = None,
      detect_faces: bool = True,
      estimate_rolling_vitals: bool = True,
      fdet_max_faces: int = 1,
      fdet_fs: float = 1.0,
      fdet_score_threshold: float = 0.9,
      fdet_iou_threshold: float = 0.3,
      export_to_json: bool = True,
      export_dir: str = "."
    ):
    """Initialisation. Loads face detection model if necessary.

    Args:
      method: The rPPG method to be used for inference.
      mode: Operate in batch or burst mode
      api_key: Usage key for the VitalLens API (required for Method.VITALLENS)
      detect_faces: `True` if faces need to be detected, otherwise `False`.
      estimate_rolling_vitals: Set `True` to compute rolling vitals (e.g., `rolling_heart_rate`).
      fdet_max_faces: The maximum number of faces to detect (if necessary).
      fdet_fs: Frequency [Hz] at which faces should be scanned. Detections are
        linearly interpolated for remaining frames.
      fdet_score_threshold: Face detection score threshold.
      fdet_iou_threshold: Face detection iou threshold.
      export_to_json: If `True`, write results to a json file.
      export_dir: The directory to which json files are written.
    """
    self.api_key = api_key
    # Load the config and model
    self.config = load_config(method.name.lower() + ".yaml")
    self.method = method
    self.mode = mode
    if self.config['model'] == 'g':
      self.rppg = GRPPGMethod(config=self.config, mode=mode)
    elif self.config['model'] == 'chrom':
      self.rppg = CHROMRPPGMethod(config=self.config, mode=mode)
    elif self.config['model'] == 'pos':
      self.rppg = POSRPPGMethod(config=self.config, mode=mode)
    elif self.config['model'] == 'vitallens':
      if self.api_key is None or self.api_key == '':
        raise ValueError("An API key is required to use Method.VITALLENS, but was not provided. "
                         "Get one for free at https://www.rouast.com/api.")
      self.rppg = VitalLensRPPGMethod(config=self.config, mode=mode, api_key=self.api_key)
    else:
      raise ValueError(f"Method {self.config['model']} not implemented!")
    self.detect_faces = detect_faces
    self.estimate_rolling_vitals = estimate_rolling_vitals
    self.export_to_json = export_to_json
    self.export_dir = export_dir
    if detect_faces:
      # Initialise face detector
      self.face_detector = FaceDetector(
        max_faces=fdet_max_faces, fs=fdet_fs, score_threshold=fdet_score_threshold,
        iou_threshold=fdet_iou_threshold)
    assert not (fdet_max_faces > 1 and mode == Mode.BURST), "burst mode only supported for one face"
  def __call__(
      self,
      video: Union[np.ndarray, str],
      faces: Union[np.ndarray, list] = None,
      fps: float = None,
      override_fps_target: float = None,
      override_global_parse: bool = None,
      export_filename: str = None
    ) -> list:
    """Run rPPG inference.

    Args:
      video: The video to analyze. Either a np.ndarray of shape (n_frames, h, w, 3)
        with a sequence of frames in unscaled uint8 RGB format, or a path to a
        video file. Note that aggressive video encoding destroys the rPPG signal.
      faces: Face boxes in flat point form, containing [x0, y0, x1, y1] coords.
        Ignored unless detect_faces=False. Pass a list or np.ndarray of
        - shape (n_faces, n_frames, 4) for multiple faces detected on multiple frames,
        - shape (n_frames, 4) for single face detected on mulitple frames, or
        - shape (4,) for a single face detected globally, or
        - `None` to assume all frames already cropped to the same single face detection.
      fps: Sampling frequency of the input video. Required if type(video) == np.ndarray. 
      override_fps_target: Target fps at which rPPG inference should be run (optional).
        If not provided, will use default of the selected method.
      override_global_parse: If True, always use global parse. If False, don't use global parse.
        If None, choose based on video.
      export_filename: Filename for json export if applicable.
    Returns:
      result: Analysis results as a list of faces in the following format:
        [
          {
            'face': {
              'coordinates': <Face coordinates for each frame as np.ndarray of shape (n_frames, 4)>,
              'confidence': <Face live confidence for each frame as np.ndarray of shape (n_frames,)>,
              'note': <Explanatory note>
            },
            'vital_signs': {
              'heart_rate': {
                'value': <Estimated value as float scalar>,
                'unit': <Value unit>,
                'confidence': <Estimation confidence as float scalar>,
                'note': <Explanatory note>
              },
              'respiratory_rate': {
                'value': <Estimated value as float scalar>,
                'unit': <Value unit>,
                'confidence': <Estimation confidence as float scalar>,
                'note': <Explanatory note>
              },
              'ppg_waveform': {
                'data': <Estimated waveform value for each frame as np.ndarray of shape (n_frames,)>,
                'unit': <Data unit>,
                'confidence': <Estimation confidence for each frame as np.ndarray of shape (n_frames,)>,
                'note': <Explanatory note>
              },
              'respiratory_waveform': {
                'data': <Estimated waveform value for each frame as np.ndarray of shape (n_frames,)>,
                'unit': <Data unit>,
                'confidence': <Estimation confidence for each frame as np.ndarray of shape (n_frames,)>,
                'note': <Explanatory note>
              },
            },
            "message": <Message about estimates>
          },
          { 
            <same structure for face 2 if present>
          },
          ...
        ]
    """
    # Probe inputs
    if self.mode == Mode.BURST and not isinstance(video, np.ndarray):
      raise ValueError("Must provide `np.ndarray` inputs for burst mode.")
    if self.mode == Mode.BURST and video.shape[0] > (API_MAX_FRAMES - self.rppg.n_inputs + 1):
      raise ValueError(f"Maximum number of frames in burst mode is {API_MAX_FRAMES - self.rppg.n_inputs + 1}, but received {video.shape[0]}.")
    inputs_shape, fps, _ = probe_image_inputs(video, fps=fps, allow_image=False)
    # TODO: Optimize performance of simple rPPG methods for long videos
    # Warning if using long video
    target_fps = override_fps_target if override_fps_target is not None else self.rppg.fps_target
    if self.method != Method.VITALLENS and inputs_shape[0]/fps*target_fps > 3600:
      logging.warning("Inference for long videos has yet to be optimized for POS / G / CHROM. This may run out of memory and crash.")
    _, height, width, _ = inputs_shape
    if self.detect_faces:
      # Detect faces
      if self.mode == Mode.BURST:
        faces_rel, _ = self.face_detector(inputs=video[-1:], n_frames=1, fps=fps)
      else:
        faces_rel, _ = self.face_detector(inputs=video, n_frames=inputs_shape[0], fps=fps)
      # If no faces detected: return empty list
      if len(faces_rel) == 0:
        logging.warning("No faces to analyze")
        return []
      # Convert to absolute units
      faces = (faces_rel * [width, height, width, height]).astype(np.int64)
      # Face axis first
      faces = np.transpose(faces, (1, 0, 2))
    # Check if the faces are valid
    faces = check_faces(faces, inputs_shape)
    # Run separately for each face
    results = []
    for face in faces:
      # Run selected rPPG method
      data, unit, conf, note, live = self.rppg(inputs=video,
                                               faces=face,
                                               fps=fps,
                                               override_fps_target=override_fps_target,
                                               override_global_parse=override_global_parse)
      # Parse face results
      face_result = {'face': {
        'coordinates': face,
        'confidence': live,
        'note': "Face detection coordinates for this face, along with live confidence levels."
      }}
      # Parse vital signs results
      vital_signs_results = {}
      for name in self.config['signals']:
        if name in data and name in unit and name in conf and name in note:
          vital_signs_results[name] = {
            f"{'data' if 'waveform' in name else 'value'}": data[name],
            'unit': unit[name],
            'confidence': conf[name],
            'note': note[name]
          }
      if self.estimate_rolling_vitals:
        try:
          if 'ppg_waveform' in self.config['signals']:
            rolling_hr = estimate_hr_from_signal(
              signal=data['ppg_waveform'],
              f_s=fps,
              window_size=CALC_HR_WINDOW_SIZE,
              scope=EScope.ROLLING,
              method=EMethod.PERIODOGRAM
            )
            rolling_conf = rolling_calc(
              x=conf['ppg_waveform'],
              calc_fn=lambda x: np.nanmean(x),
              min_window_size=CALC_HR_WINDOW_SIZE,
              max_window_size=CALC_HR_WINDOW_SIZE,
              overlap=CALC_HR_WINDOW_SIZE//2
            )
            vital_signs_results['rolling_heart_rate'] = {
              'data': rolling_hr,
              'unit': 'bpm',
              'confidence': rolling_conf,
              'note': 'Estimate of the rolling heart rate using VitalLens, along with frame-wise confidences between 0 and 1.',
            }
          if 'respiratory_waveform' in self.config['signals']:
            rolling_rr = estimate_rr_from_signal(
              signal=data['respiratory_waveform'],
              f_s=fps,
              window_size=CALC_RR_WINDOW_SIZE,
              scope=EScope.ROLLING,
              method=EMethod.PERIODOGRAM
            )
            rolling_conf = rolling_calc(
              x=conf['respiratory_waveform'],
              calc_fn=lambda x: np.nanmean(x),
              min_window_size=CALC_RR_WINDOW_SIZE,
              max_window_size=CALC_RR_WINDOW_SIZE,
              overlap=CALC_RR_WINDOW_SIZE//2
            )
            vital_signs_results['rolling_respiratory_rate'] = {
              'data': rolling_rr,
              'unit': 'bpm',
              'confidence': rolling_conf,
              'note': 'Estimate of the rolling respiratory rate using VitalLens, along with frame-wise confidences between 0 and 1.',
            }
        except ValueError as e:
          logging.debug(f"Issue while computing rolling vitals: {e}")
      face_result['vital_signs'] = vital_signs_results
      face_result['message'] = DISCLAIMER
      results.append(face_result)
    # Export to json
    if self.export_to_json:
      os.makedirs(self.export_dir, exist_ok=True)
      export_filename = f"{export_filename}.json" if export_filename is not None else f"vitallens_{datetime.now().strftime('%Y%m%d_%H%M%S')}.json"
      with open(os.path.join(self.export_dir, export_filename), 'w') as f:
        json.dump(convert_ndarray_to_list(results), f, indent=4)
    return results
  def reset(self):
    """Reset"""
    self.rppg.reset()



# Copyright (c) 2024 Philipp Rouast
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
import pytest

import sys
sys.path.append('../vitallens-python')

from vitallens.ssd import nms, enforce_temporal_consistency
from vitallens.ssd import interpolate_unscanned_frames, FaceDetector

@pytest.mark.parametrize("max_output_size", [1, 2, 3])
def test_nms(max_output_size):
  boxes = np.asarray([[[0.0, 0.0, 0.25, 0.25],
                       [0.5, 0.0, 1.0, 0.5],
                       [0.5, 0.125, 0.875, 0.5],
                       [0.25, 0.5, 0.5, 0.75],
                       [0.375, 0.625, 0.625, 0.875],
                       [0.5, 0.75, 0.75, 1.0]],
                      [[0.0, 0.0, 0.25, 0.25],
                       [0.5, 0.0, 1.0, 0.5],
                       [0.5, 0.125, 0.875, 0.5],
                       [0.25, 0.5, 0.5, 0.75],
                       [0.375, 0.625, 0.625, 0.875],
                       [0.5, 0.75, 0.75, 1.0]]])
  scores = np.asarray([[0.01,
                        0.99,
                        0.95,
                        0.97,
                        0.80,
                        0.09],
                       [0.01,
                        0.99,
                        0.95,
                        0.97,
                        0.80,
                        0.98]])
  if max_output_size == 1:
    idxs = np.asarray([[1], [1]])
    num_valid = np.asarray([1, 1])
  elif max_output_size == 2:
    idxs = np.asarray([[1, 3], [1, 5]])
    num_valid = np.asarray([2, 2])
  else:
    idxs = np.asarray([[1, 3, 0], [1, 5, 3]])
    num_valid = np.asarray([2, 3])
  out_idxs, out_num_valid = nms(boxes=boxes, scores=scores, max_output_size=max_output_size, iou_threshold=0.45, score_threshold=0.9)
  np.testing.assert_equal(
    out_idxs,
    idxs)
  np.testing.assert_equal(
    out_num_valid,
    num_valid)

def test_enforce_temporal_consistency():
  # Example with 2 moving faces, 5 time steps, no detection for face 1 in time step 2, faces swapped in time step 4
  boxes = np.array(
    [[[.125, .125, .375, .375], [.125, .5, .375, .75]],
     [[.125, .25,  .375, .5  ], [.25,  .5, .5,   .75]],
     [[.125, .375, .375, .625], [.375, .5, .625, .75]],
     [[.5,   .5,   .75,  .75 ], [.125, .5, .375, .75]],
     [[.125, .625, .375, .875], [.625, .5, .875, .75]]])
  info = np.array(
    [[[0, 1, 1, .99], [0, 1, 1, .99]],
     [[1, 1, 0, .2 ], [1, 1, 1, .99]],
     [[2, 1, 1, .99], [2, 1, 1, .99]],
     [[3, 1, 1, .99], [3, 1, 1, .99]],
     [[4, 1, 1, .99], [4, 1, 1, .99]]])
  boxes_out, info_out = enforce_temporal_consistency(
    boxes=boxes, info=info, n_frames=5)
  np.testing.assert_equal(
    boxes_out,
      np.array(
        [[[.125, .5, .375, .75], [.125, .125, .375, .375]],
         [[.25,  .5, .5,   .75], [.125, .25,  .375, .5  ]],
         [[.375, .5, .625, .75], [.125, .375, .375, .625]],
         [[.5,   .5, .75,  .75], [.125, .5,   .375, .75 ]],
         [[.625, .5, .875, .75], [.125, .625, .375, .875]]]))
  np.testing.assert_equal(
    info_out,
      np.array(
        [[[0, 1, 1, .99], [0, 1, 1, .99]],
         [[1, 1, 1, .99], [1, 1, 0, .2 ]],
         [[2, 1, 1, .99], [2, 1, 1, .99]],
         [[3, 1, 1, .99], [3, 1, 1, .99]],
         [[4, 1, 1, .99], [4, 1, 1, .99]]]))
  
def test_interpolate_unscanned_frames():
  # Example with 2 moving faces, 3 time steps, no detection for face 1 in time step 2, faces swapped in time step 4
  boxes = np.array(
    [[[.125, .5, .375, .75], [.125, .125, .375, .375]],
     [[.25,  .5, .5,   .75], [.125, .25,  .375, .5  ]],
     [[.375, .5, .625, .75], [.125, .375, .375, .625]]])
  info = np.array(
    [[[0, 1, 1, .99], [0, 1, 1, .99]],
     [[2, 1, 1, .99], [2, 1, 0, .2 ]],
     [[4, 1, 1, .99], [4, 1, 1, .99]]])
  boxes_out, info_out = interpolate_unscanned_frames(
    boxes=boxes, info=info, n_frames=5)
  np.testing.assert_equal(
    boxes_out,
      np.array(
        [[[.125,  .5, .375,  .75], [.125, .125,  .375, .375 ]],
         [[.1875, .5, .4375, .75], [.125, .1875, .375, .4375]],
         [[.25,   .5, .5,    .75], [.125, .25,   .375, .5   ]],
         [[.3125, .5, .5625, .75], [.125, .3125, .375, .5625]],
         [[.375,  .5, .625,  .75], [.125, .375,  .375, .625 ]]]))
  np.testing.assert_equal(
    info_out,
      np.array(
        [[[0, 1, 1, .99], [0, 1, 1, .99]],
         [[1, 0, 0, 0  ], [1, 0, 0, 0  ]], # Imperfection of the implementation
         [[2, 1, 1, .99], [2, 1, 0, .2 ]],
         [[3, 0, 0, 0  ], [3, 0, 0, 0  ]],
         [[4, 1, 1, .99], [4, 1, 1, .99]]]))

@pytest.mark.parametrize("file", [True, False])
def test_FaceDetector(request, file):
  det = FaceDetector(
    max_faces=2, fs=1.0, iou_threshold=0.45, score_threshold=0.9)
  if file:
    test_video_path = request.getfixturevalue('test_video_path')
    test_video_shape = request.getfixturevalue('test_video_shape')
    test_video_fps = request.getfixturevalue('test_video_fps')
    boxes, info = det(inputs=test_video_path,
                      n_frames=test_video_shape[0],
                      fps=test_video_fps)
  else:
    test_video_ndarray = request.getfixturevalue('test_video_ndarray')
    test_video_fps = request.getfixturevalue('test_video_fps')
    boxes, info = det(inputs=test_video_ndarray,
                      n_frames=test_video_ndarray.shape[0],
                      fps=test_video_fps)
  assert boxes.shape == (360, 1, 4)
  assert info.shape == (360, 1, 4)
  np.testing.assert_allclose(boxes[0,0],
                             [0.32223, 0.118318, 0.572684, 0.696835],
                             atol=0.01)
  


import argparse
import concurrent.futures
import cv2
import numpy as np
from prpy.constants import SECONDS_PER_MINUTE
from prpy.numpy.face import get_upper_body_roi_from_det
from prpy.numpy.physio import estimate_rate_from_signal, EScope, EMethod
from prpy.numpy.physio import HR_MIN, HR_MAX, RR_MIN, RR_MAX
import sys
import threading
import time
import warnings

sys.path.append('../vitallens-python')
from vitallens import VitalLens, Mode, Method
from vitallens.buffer import SignalBuffer, MultiSignalBuffer
from vitallens.constants import API_MIN_FRAMES

def draw_roi(frame, roi):
  roi = np.asarray(roi).astype(np.int32)
  frame = cv2.rectangle(frame, (roi[0], roi[1]), (roi[2], roi[3]), (0, 255, 0), 1)

def draw_signal(frame, roi, sig, sig_name, sig_conf_name, draw_area_tl_x, draw_area_tl_y, color):
  def _draw(frame, vals, display_height, display_width, min_val, max_val, color, thickness):
    height_mult = display_height/(max_val - min_val)
    width_mult = display_width/(vals.shape[0] - 1)
    p1 = (int(draw_area_tl_x), int(draw_area_tl_y + (max_val - vals[0]) * height_mult))
    for i, s in zip(range(1, len(vals)), vals[1:]):
      p2 = (int(draw_area_tl_x + i * width_mult), int(draw_area_tl_y + (max_val - s) * height_mult))
      frame = cv2.line(frame, p1, p2, color, thickness)
      p1 = p2
  # Derive dims from roi
  display_height = (roi[3] - roi[1]) / 2.0
  display_width = (roi[2] - roi[0]) * 1.2
  # Draw signal
  if sig_name in sig:
    vals = np.asarray(sig[sig_name])
    min_val = np.min(vals)
    max_val = np.max(vals)
    if max_val - min_val == 0:
      return frame
    _draw(frame=frame, vals=vals, display_height=display_height, display_width=display_width,
          min_val=min_val, max_val=max_val, color=color, thickness=2)
  # Draw confidence
  if sig_conf_name in sig:
    vals = np.asarray(sig[sig_conf_name])
    _draw(frame=frame, vals=vals, display_height=display_height, display_width=display_width,
          min_val=0., max_val=1., color=color, thickness=1)

def draw_fps(frame, fps, text, draw_area_bl_x, draw_area_bl_y):
  cv2.putText(frame, text=f"{text}: {fps:.1f}", org=(draw_area_bl_x, draw_area_bl_y),
    fontFace=cv2.FONT_HERSHEY_SIMPLEX, fontScale=0.6, color=(0,255,0), thickness=1)

def draw_vital(frame, sig, text, sig_name, fps, color, draw_area_bl_x, draw_area_bl_y):
  if sig_name in sig:
    f_range = (HR_MIN/SECONDS_PER_MINUTE, HR_MAX/SECONDS_PER_MINUTE) if 'ppg' in sig_name else (RR_MIN/SECONDS_PER_MINUTE, RR_MAX/SECONDS_PER_MINUTE)
    val = estimate_rate_from_signal(signal=sig[sig_name], f_s=fps, f_range=f_range, scope=EScope.GLOBAL, method=EMethod.PERIODOGRAM)
    cv2.putText(frame, text=f"{text}: {val:.1f}", org=(draw_area_bl_x, draw_area_bl_y),
      fontFace=cv2.FONT_HERSHEY_SIMPLEX, fontScale=0.6, color=color, thickness=2)

class VitalLensRunnable:
  def __init__(self, method, api_key):
    self.active = threading.Event()
    self.result = []
    self.vl = VitalLens(method=method,
                        mode=Mode.BURST,
                        api_key=api_key,
                        detect_faces=True,
                        estimate_rolling_vitals=True,
                        export_to_json=False)
  def __call__(self, inputs, fps):
    self.active.set()
    self.result = self.vl(np.asarray(inputs), fps=fps)
    self.active.clear()

def run(args):
  cap = cv2.VideoCapture(0)
  executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
  vl = VitalLensRunnable(method=args.method, api_key=args.api_key)
  signal_buffer = MultiSignalBuffer(size=240, ndim=1, ignore_k=['face'])
  fps_buffer = SignalBuffer(size=240, ndim=1, pad_val=np.nan)
  frame_buffer = []
  # Sample frames from cv2 video stream attempting to achieve this framerate
  target_fps = 30.
  # Check if the webcam is opened correctly
  if not cap.isOpened():
    raise IOError("Cannot open webcam")
  # Read first frame to get dims
  _, frame = cap.read()
  height, width, _ = frame.shape
  roi = None
  i = 0
  t, p_t = time.time(), time.time()
  fps, p_fps = 30.0, 30.0
  ds_factor = 1
  n_frames = 0
  signals = None
  while True:
    ret, frame = cap.read()
    # Measure frequency
    t_prev = t
    t = time.time()
    if not vl.active.is_set():
      # Process result if available
      if len(vl.result) > 0:
        # Results are available - fetch and reset
        result = vl.result[0]
        vl.result = []
        # Update the buffer
        signals = signal_buffer.update({
          **{
            f"{key}_sig": value['value'] if 'value' in value else np.array(value['data'])
            for key, value in result['vital_signs'].items()
          },
          **{
            f"{key}_conf": value['confidence'] if isinstance(value['confidence'], np.ndarray) else np.array(value['confidence'])
            for key, value in result['vital_signs'].items()
          },
          'face_conf': result['face']['confidence'],
        }, dt=n_frames)
        with warnings.catch_warnings():
          warnings.simplefilter("ignore", category=RuntimeWarning)
          # Measure actual effective sampling frequency at which neural net input was sampled
          fps = np.nanmean(fps_buffer.update([(1./(t - t_prev))/ds_factor], dt=n_frames))
        roi = get_upper_body_roi_from_det(result['face']['coordinates'][-1], clip_dims=(width, height), cropped=True)
        # Measure prediction frequency - how often predictions are made
        p_t_prev = p_t
        p_t = time.time()
        p_fps = 1./(p_t - p_t_prev)
      else:
        # No results available
        roi = None
        signal_buffer.clear()
      # Start next prediction
      if len(frame_buffer) >= (API_MIN_FRAMES if args.method == Method.VITALLENS else 1):
        n_frames = len(frame_buffer)
        executor.submit(vl, frame_buffer.copy(), fps)
        frame_buffer.clear()
    # Sample frames
    if i % ds_factor == 0:
      # Add current frame to the buffer (BGR -> RGB)
      frame_buffer.append(frame[...,::-1])
    i += 1
    # Display
    if roi is not None:
      draw_roi(frame, roi)
      draw_signal(
        frame=frame, roi=roi, sig=signals, sig_name='ppg_waveform_sig', sig_conf_name='ppg_waveform_conf',
        draw_area_tl_x=roi[2]+20, draw_area_tl_y=roi[1], color=(0, 0, 255))
      draw_signal(
        frame=frame, roi=roi, sig=signals, sig_name='respiratory_waveform_sig', sig_conf_name='respiratory_waveform_conf',
        draw_area_tl_x=roi[2]+20, draw_area_tl_y=int(roi[1]+(roi[3]-roi[1])/2.0), color=(255, 0, 0))
      draw_fps(frame, fps=fps, text="fps", draw_area_bl_x=roi[0], draw_area_bl_y=roi[3]+20)
      draw_fps(frame, fps=p_fps, text="p_fps", draw_area_bl_x=int(roi[0]+0.4*(roi[2]-roi[0])), draw_area_bl_y=roi[3]+20)
      draw_vital(frame, sig=signals, text="hr [bpm]", sig_name='ppg_waveform_sig', fps=fps, color=(0,0,255), draw_area_bl_x=roi[2]+20, draw_area_bl_y=int(roi[1]+(roi[3]-roi[1])/2.0))
      draw_vital(frame, sig=signals, text="rr [rpm]", sig_name='respiratory_waveform_sig', fps=fps, color=(255,0,0), draw_area_bl_x=roi[2]+20, draw_area_bl_y=roi[3])
    cv2.imshow('Live', frame)
    c = cv2.waitKey(1)
    if c == 27:
      break
    # Even out fps
    dt_req = 1./target_fps - (time.time() - t)
    if dt_req > 0: time.sleep(dt_req)

  cap.release()
  cv2.destroyAllWindows()

def method_type(name):
  try:
    return Method[name]
  except KeyError:
    raise argparse.ArgumentTypeError(f"{name} is not a valid Method")

if __name__ == '__main__':
  parser = argparse.ArgumentParser()
  parser.add_argument('--api_key', type=str, default='', help='Your API key. Get one for free at https://www.rouast.com/api.')
  parser.add_argument('--method', type=method_type, default='VITALLENS', help='Choice of method (VITALLENS, POS, CHROM, or G)')
  args = parser.parse_args()
  run(args)



# Copyright (c) 2024 Rouast Labs
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
from prpy.numpy.physio import EScope, EMethod
from prpy.numpy.physio import estimate_hr_from_signal, estimate_rr_from_signal
from typing import Tuple

def reassemble_from_windows(
    x: np.ndarray,
    idxs: np.ndarray
  ) -> Tuple[np.ndarray, np.ndarray]:
  """Reassemble windowed data using corresponding idxs.

  Args:
    x: Data generated using a windowing operation. Shape (n_windows, n, window_size)
    idxs: Indices of x in the original 1-d array. Shape (n_windows, window_size)
  Returns:
    Tuple of
     - out: Reassembled data. Shape (n, n_idxs)
     - idxs: Reassembled idxs. Shape (n_idxs,)
  """
  x = np.asarray(x)
  idxs = np.asarray(idxs)
  # Transpose x (n, n_windows, window_size)
  x = np.transpose(x, (1, 0, 2))
  # Adjust indices based on their window position
  offset_idxs = idxs - np.arange(idxs.shape[0])[:, np.newaxis]
  # Find strictly increasing indices using np.maximum.accumulate
  flat_offset_idxs = offset_idxs.flatten()
  max_so_far = np.maximum.accumulate(flat_offset_idxs.flatten())
  mask = (flat_offset_idxs == max_so_far)  # Mask to keep only strictly increasing indices
  # Filter data based on mask and extract the final result values
  result = x.reshape(x.shape[0], -1)[:,mask]
  idxs = idxs.flatten()[mask]
  return result, idxs

def assemble_results(
    sig: np.ndarray,
    conf: np.ndarray,
    live: np.ndarray,
    fps: float,
    train_sig_names: list,
    pred_signals: list,
    method_name: str,
    can_provide_confidence: bool = True,
    min_t_hr: float = 2.,
    min_t_rr: float = 4.
  ) -> Tuple[dict, dict, dict, dict, np.ndarray]:
  """Assemble rPPG method results in the format expected by the API.
  Args:
    sig: The estimated signals. Shape (n_sig, n_frames)
    conf: The estimation confidence. Shape (n_sig, n_frames)
    live: The liveness confidence. Shape (n_frames,)
    fps: The sampling rate
    train_sig_names: The train signal names of the method
    pred_signals: The pred signals specs of the method
    method_name: The name of the method
    can_provide_confidence: Whether the method can provide a confidence estimate
    min_t_hr: Minimum amount of time signal to estimate hr
    min_t_rr: Minimum amount of time signal to estimate rr
  Returns:
    Tuple of
       - out_data: The estimated data/value for each signal.
       - out_unit: The estimation unit for each signal.
       - out_conf: The estimation confidence for each signal.
       - out_note: An explanatory note for each signal.
       - live: The face live confidence. Shape (n_frames,)
  """
  # Infer the signal length in seconds
  sig_t = sig.shape[1] / fps
  # Get the names of signals model outputs
  out_data, out_unit, out_conf, out_note = {}, {}, {}, {}
  confidence_note_scalar = ', along with a confidence level between 0 and 1.' if can_provide_confidence else '. This method is not capable of providing a confidence estimate, hence returning 1.'
  confidence_note_data = ', along with frame-wise confidences between 0 and 1.' if can_provide_confidence else '. This method is not capable of providing a confidence estimate, hence returning 1.'
  for name in pred_signals:
    if name == 'heart_rate' and 'ppg_waveform' in train_sig_names and sig_t > min_t_hr:
      ppg_ir_idx = train_sig_names.index('ppg_waveform')
      out_data[name] = estimate_hr_from_signal(signal=sig[ppg_ir_idx],
                                               f_s=fps,
                                               scope=EScope.GLOBAL,
                                               method=EMethod.PERIODOGRAM)
      out_unit[name] = 'bpm'
      out_conf[name] = float(np.mean(conf[ppg_ir_idx]))
      out_note[name] = f'Estimate of the global heart rate using {method_name}{confidence_note_scalar}'
    elif name == 'respiratory_rate' and 'respiratory_waveform' in train_sig_names and sig_t > min_t_rr:
      resp_idx = train_sig_names.index('respiratory_waveform')
      out_data[name] = estimate_rr_from_signal(signal=sig[resp_idx],
                                               f_s=fps,
                                               scope=EScope.GLOBAL,
                                               method=EMethod.PERIODOGRAM)
      out_unit[name] = 'bpm'
      out_conf[name] = float(np.mean(conf[resp_idx]))
      out_note[name] = f'Estimate of the global respiratory rate using {method_name}{confidence_note_scalar}'
    elif name == 'ppg_waveform':
      ppg_ir_idx = train_sig_names.index('ppg_waveform')
      out_data[name] = sig[ppg_ir_idx]
      out_unit[name] = 'unitless'
      out_conf[name] = conf[ppg_ir_idx]
      out_note[name] = f'Estimate of the ppg waveform using {method_name}{confidence_note_data}'
    elif name == 'respiratory_waveform':
      resp_idx = train_sig_names.index('respiratory_waveform')
      out_data[name] = sig[resp_idx]
      out_unit[name] = 'unitless'
      out_conf[name] = conf[resp_idx]
      out_note[name] = f'Estimate of the respiratory waveform using {method_name}{confidence_note_data}'
  return out_data, out_unit, out_conf, out_note, live



# Copyright (c) 2024 Rouast Labs
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import abc

from vitallens.enums import Mode

class RPPGMethod(metaclass=abc.ABCMeta):
  """Abstract superclass for rPPG methods"""
  def __init__(
      self,
      config: dict,
      mode: Mode
    ):
    """Initialize the `RPPGMethod`
    
    Args:
      config: The configuration dict
      mode: The operation mode
    """
    self.fps_target = config['fps_target']
    self.op_mode = mode
    self.n_inputs = 1
    self.est_window_length = config['est_window_length']
    self.est_window_overlap = config['est_window_overlap']
    self.est_window_flexible = self.est_window_length == 0
  @abc.abstractmethod
  def __call__(self, frames, faces, fps, override_fps_target, override_global_parse):
    """Run inference. Abstract method to be implemented in subclasses."""
    pass
  @abc.abstractmethod
  def reset(self):
    """Reset. Abstract method to be implemented in subclasses."""
    pass
  


# Copyright (c) 2024 Rouast Labs
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
from typing import Union
import warnings

class SignalBuffer:
  """A buffer for an arbitrary float signal.
  - Builds the buffer up over time.
  - Returns overlaps using mean.
  """
  def __init__(
      self,
      size: int,
      ndim: int,
      pad_val: float = 0,
      init_t: int = 0
    ):
    """Initialise the signal buffer.
    Args:
      size: The temporal length of the buffer. Signals are pushed through in FIFO order.
      ndim: The number of dims for the signal (e.g., ndim=1 means scalar signal with time dimension)
      pad_val: Constant scalar to pad empty time steps with
      init_t: Time for initialisation
    """
    self.size = size
    self.pad_val = pad_val
    self.min_increment = 1
    self.ndim = ndim
    # Buffer is a nested list with up to self.size lists
    # Each element is a tuple
    # - 0: t_start
    # - 1: t_end
    # - 2: signal  
    self.buffer = []
    self.t = init_t
    self.out = None
  def update(
      self,
      signal: Union[list, np.ndarray],
      dt: int
    ) -> np.ndarray:
    """Update the buffer with new signal series and amount of time steps passed.
    - Support arbitrary number of signal dims, but mostly intended for scalar series or one more dim
    (e.g., RGB signal over time).
    Args:
      signal: The signal. list or ndarray, shape (n_frames, dim1, dim2, ...)
      dt: The number of time steps passed. Scalar
    Returns:
      out: The signal of buffer size, with overlaps averaged. Shape (self.size, dim1, dim2, ...)
    """
    if isinstance(signal, list): signal = np.asarray(signal)
    if signal.size == 1 and self.ndim == 1: signal = np.full((dt,), fill_value=signal)
    assert isinstance(signal, np.ndarray), f"signal should be np.ndarray but is {type(signal)}"
    assert len(signal.shape) == self.ndim
    assert len(signal) >= 1
    assert dt >= self.min_increment
    # Initialise self.out if necessary
    if self.out is None:
      self.out = np.empty((self.size, self.size,) + signal.shape[1:], signal.dtype)
    self.out[:] = np.nan
    # Update self.t
    self.t += dt
    # Update self.buffer
    self.buffer.append((self.t - signal.shape[0], self.t, signal))
    # Delete old buffer elements
    i = 0
    while i < len(self.buffer):
      if self.buffer[i][1] <= self.t - self.size:
        self.buffer.pop(0)
      else:
        i += 1
    return self.get()
  def get(self) -> np.ndarray:
    """Get the series of current buffer contents, with overlaps averaged.
    Returns:
      out: The signal of buffer size, with overlaps averaged. Shape (self.size, dim1, dim2, ...)
    """
    # No elements yet
    assert self.t > 0, "Update at least once before calling get()"
    # Assign buffer elements to self.out
    for i in range(len(self.buffer)):
      adj_t = self.t - self.size
      adj_t_start = self.buffer[i][0] - adj_t
      adj_t_end = self.buffer[i][1] - adj_t
      outside = 0 if adj_t_start >= 0 else abs(adj_t_start)
      self.out[i][adj_t_start+outside:adj_t_end] = self.buffer[i][2][outside:]
    # Reduce via mean (ignore warnings due to nan)
    with warnings.catch_warnings():
      warnings.simplefilter("ignore", category=RuntimeWarning)
      result = np.nanmean(self.out, axis=0)
    # Replace np.nan with pad_val
    return np.nan_to_num(result, nan=self.pad_val)
  def clear(self):
    """Clear the contents"""
    self.buffer.clear()
    self.t = 0
    self.out = None

class MultiSignalBuffer:
  """Manages a dict of SignalBuffer instances"""
  def __init__(
      self,
      size: int,
      ndim: int,
      ignore_k: list,
      pad_val: float = 0
    ):
    """Initialise the multi signal buffer.
    Args:
      size: The temporal length of each buffer. Signals are pushed through in FIFO order.
      ndim: The number of dims for each signal (e.g., ndim=1 means scalar signal with time dimension)
      ignore_k: List of keys to ignore in update step
      pad_val: Constant scalar to pad empty time steps with
    """
    self.size = size
    self.ndim = ndim
    self.min_increment = 1
    self.ignore_k = ignore_k
    self.pad_val = pad_val
    self.signals = {}
    self.t = 0
  def update(
      self,
      signals: dict,
      dt: int
    ) -> dict:
    """Initialise or update each of the buffers corresponding to the entries in signals dict.
    Args:
      signals: Dictionary of signal updates. Each entry ndarray or list of shape (n_frames, dim1, dim2, ...)
      dt: The number of time steps passed. Scalar
    Returns:
      out: Dictionary of buffered signals, with overlaps averaged. Each entry of shape (self.size, dim1, dim2, ...)
    """
    assert isinstance(signals, dict)
    result = {}
    for k in signals:
      if k in self.ignore_k:
        continue
      if k not in self.signals:
        # Add k to self.signals
        self.signals[k] = SignalBuffer(
          size=self.size, ndim=self.ndim, pad_val=self.pad_val, init_t=self.t)
      # Run update
      result[k] = self.signals[k].update(signals[k], dt)
    self.t += dt
    return result
  def get(self) -> dict:
    """Get the series of current buffer contents, with overlaps averaged.
    Returns:
      out: Dictionary of buffered signals, with overlaps averaged. Each entry of shape (self.size, dim1, dim2, ...)
    """
    assert self.t > 0, "Update at least once before calling get()"
    result = {}
    for k in self.signals:
      if k in self.ignore_k:
        continue
      result[k] = self.signals[k].get()
    return result
  def clear(self):
    """Clear the contents"""
    for k in self.signals: 
      self.signals[k].clear()
    self.signals = {}
    


# Copyright (c) 2024 Rouast Labs
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import abc
import numpy as np
from prpy.numpy.face import get_roi_from_det
from prpy.numpy.image import reduce_roi, parse_image_inputs
from prpy.numpy.interp import interpolate_filtered
from typing import Union, Tuple

from vitallens.buffer import SignalBuffer
from vitallens.constants import CALC_HR_MIN_WINDOW_SIZE
from vitallens.enums import Mode
from vitallens.methods.rppg_method import RPPGMethod
from vitallens.signal import assemble_results
from vitallens.utils import merge_faces

class SimpleRPPGMethod(RPPGMethod):
  """A simple rPPG method using a handcrafted algorithm based on RGB signal trace"""
  def __init__(
      self,
      config: dict,
      mode: Mode
    ):
    """Initialize the `SimpleRPPGMethod`
    
    Args:
      config: The configuration dict
      mode: The operation mode
    """
    super(SimpleRPPGMethod, self).__init__(config=config, mode=mode)
    self.model = config['model']
    self.roi_method = config['roi_method']
    self.signals = config['signals']
    if mode == Mode.BURST:
      self.buffer = SignalBuffer(size=self.est_window_length, ndim=2)
  @abc.abstractmethod
  def algorithm(
      self,
      rgb: np.ndarray,
      fps: float
    ):
    """The algorithm. Abstract method to be implemented by subclasses."""
    pass
  @abc.abstractmethod
  def pulse_filter(self, 
      sig: np.ndarray,
      fps: float
    ) -> np.ndarray:
    """The post-processing filter to be applied to estimated pulse signal. Abstract method to be implemented by subclasses."""
    pass
  def __call__(
      self,
      inputs: Union[np.ndarray, str],
      faces: np.ndarray,
      fps: float,
      override_fps_target: float = None,
      override_global_parse: float = None,
    ) -> Tuple[dict, dict, dict, dict, np.ndarray]:
    """Estimate pulse signal from video frames using the subclass algorithm.

    Args:
      inputs: The video to analyze. Either a np.ndarray of shape (n_frames, h, w, 3)
        in unscaled uint8 RGB format, or a path to a video file.
      faces: The face detection boxes as np.int64. Shape (n_frames, 4) in form (x0, y0, x1, y1)
      fps: The rate at which video was sampled.
      override_fps_target: Override the method's default inference fps (optional).
      override_global_parse: Has no effect here.
    Returns:
      Tuple of
       - data: A dictionary with the values of the estimated vital signs.
       - unit: A dictionary with the units of the estimated vital signs.
       - conf: A dictionary with the confidences of the estimated vital signs.
       - note: A dictionary with notes on the estimated vital signs.
       - live: Dummy live confidence estimation (set to always 1). Shape (n_frames,)
    """
    # Compute temporal union of ROIs
    u_roi = merge_faces(faces)
    faces = faces - [u_roi[0], u_roi[1], u_roi[0], u_roi[1]]
    # Parse the inputs
    frames_ds, fps, inputs_shape, ds_factor, _ = parse_image_inputs(
      inputs=inputs, fps=fps, roi=u_roi, target_size=None,
      target_fps=override_fps_target if override_fps_target is not None else self.fps_target,
      preserve_aspect_ratio=False, library='prpy', scale_algorithm='bilinear', 
      trim=None, allow_image=False, videodims=True)
    assert inputs_shape[0] == faces.shape[0], "Need same number of frames as face detections"
    faces_ds = faces[0::ds_factor]
    assert frames_ds.shape[0] == faces_ds.shape[0], "Need same number of frames as face detections"
    fps_ds = fps*1.0/ds_factor
    # Extract rgb signal (n_frames_ds, 3)
    if self.op_mode == Mode.BATCH:
      roi_ds = np.asarray([get_roi_from_det(f, roi_method=self.roi_method) for f in faces_ds], dtype=np.int64) # roi for each frame (n, 4)
      rgb_ds = reduce_roi(video=frames_ds, roi=roi_ds)
    else:
      # Use the last face detection for cropping (n_frames, 3)
      rgb_ds = reduce_roi(video=frames_ds, roi=np.asarray(get_roi_from_det(faces_ds[-1], roi_method=self.roi_method), dtype=np.int64))
      # Push to buffer and get buffered vals (pred_window_length, 3)
      rgb_ds = self.buffer.update(rgb_ds, dt=inputs.shape[0])
    # Perform rppg algorithm step (n_frames_ds,)
    sig_ds = self.algorithm(rgb_ds, fps_ds)
    # Interpolate to original sampling rate (n_frames,)
    sig = interpolate_filtered(t_in=np.arange(inputs_shape[0])[0::ds_factor],
                               s_in=sig_ds,
                               t_out=np.arange(inputs_shape[0]),
                               axis=1, extrapolate=True)
    # Filter and add dim (1, n_frames)
    sig = self.pulse_filter(sig, fps)
    sig = np.expand_dims(sig, axis=0)
    # Simple rPPG method cannot specify confidence or live. Set to always 1.
    conf = np.ones_like(sig)
    live = np.ones_like(sig[0])
    # Assemble and return the results
    return assemble_results(sig=sig,
                            conf=conf,
                            live=live,
                            fps=fps,
                            train_sig_names=['ppg_waveform'],
                            pred_signals=self.signals,
                            method_name=self.model,
                            min_t_hr=CALC_HR_MIN_WINDOW_SIZE,
                            can_provide_confidence=False)
  def reset(self):
    """Reset"""
    if self.op_mode == Mode.BURST:
      self.buffer.clear()



# Copyright (c) 2024 Rouast Labs
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
import pytest

import sys
sys.path.append('../vitallens-python')

from vitallens.buffer import SignalBuffer, MultiSignalBuffer

@pytest.mark.parametrize("pad_val", [0, -1])
def test_signal_buffer(pad_val):
  # 1 dim
  buffer = SignalBuffer(size=8, ndim=1, pad_val=pad_val)
  with pytest.raises(Exception):
    buffer.get()
  np.testing.assert_allclose(
    buffer.update(signal=[.2, .4,], dt=2),
    np.asarray([pad_val, pad_val, pad_val, pad_val, pad_val, pad_val, .2, .4]))
  np.testing.assert_allclose(
    buffer.update(signal=[.1, .3, .5, .6], dt=2),
    np.asarray([pad_val, pad_val, pad_val, pad_val, .15, .35, .5, .6]))
  np.testing.assert_allclose(
    buffer.update(signal=[.6, .7], dt=1),
    np.asarray([pad_val, pad_val, pad_val, .15, .35, .5, .6, .7]))
  np.testing.assert_allclose(
    buffer.update(signal=[.8, .9, .8, .7, .6], dt=4),
    np.asarray([.35, .5, .6, .75, .9, .8, .7, .6]))
  # 2 dim
  buffer = SignalBuffer(size=4, ndim=2, pad_val=pad_val)
  with pytest.raises(Exception):
    buffer.get()
  np.testing.assert_allclose(
    buffer.update(signal=[[.1, .2,]], dt=1),
    np.asarray([[pad_val, pad_val], [pad_val, pad_val], [pad_val, pad_val], [.1, .2]]))
  np.testing.assert_allclose(
    buffer.update(signal=[[.1, .3], [.5, .6]], dt=3),
    np.asarray([[.1, .2], [pad_val, pad_val], [.1, .3], [.5, .6]]))
  np.testing.assert_allclose(
    buffer.update(signal=[[.3, .2], [.5, .6], [.5, .6]], dt=2),
    np.asarray([[.1, .3], [.4, .4], [.5, .6], [.5, .6]]))

@pytest.mark.parametrize("pad_val", [0, -1])
def test_multi_signal_buffer(pad_val):
  # 1 dim, 2 signals
  buffer = MultiSignalBuffer(size=8, ndim=1, ignore_k=[], pad_val=pad_val)
  with pytest.raises(Exception):
    buffer.get()
  with pytest.raises(Exception):
    buffer.update(signals=[0., 1.])
  out = buffer.update(
    signals={"a": [.2, .4,], "b": [.1, .2]}, dt=1)
  np.testing.assert_allclose(
    out["a"],
    np.asarray([pad_val, pad_val, pad_val, pad_val, pad_val, pad_val, .2, .4]))
  np.testing.assert_allclose(
    out["b"],
    np.asarray([pad_val, pad_val, pad_val, pad_val, pad_val, pad_val, .1, .2]))
  out = buffer.update(
    signals={"a": [.1, .3, .5, .6], "b": [.1, .4, .5]}, dt=2)
  np.testing.assert_allclose(
    out["a"],
    np.asarray([pad_val, pad_val, pad_val, pad_val, .15, .35, .5, .6]))
  np.testing.assert_allclose(
    out["b"],
    np.asarray([pad_val, pad_val, pad_val, pad_val, .1, .15, .4, .5]))
  out = buffer.update(
    signals={"a": [.6, .7], "b": [.3], "c": [.4, .8]}, dt=1)
  np.testing.assert_allclose(
    out["a"],
    np.asarray([pad_val, pad_val, pad_val, .15, .35, .5, .6, .7]))
  np.testing.assert_allclose(
    out["b"],
    np.asarray([pad_val, pad_val, pad_val, .1, .15, .4, .5, .3]))
  np.testing.assert_allclose(
    out["c"],
    np.asarray([pad_val, pad_val, pad_val, pad_val, pad_val, pad_val, .4, .8]))
  # 2 dim, 2 signals
  buffer = MultiSignalBuffer(size=4, ndim=2, pad_val=pad_val, ignore_k=['c'])
  out = buffer.update(
    signals={"a": [[.2, .4,], [.1, .4]], "b": [[.1, .2], [.6, .6]]}, dt=1)
  np.testing.assert_allclose(
    out["a"],
    np.asarray([[pad_val, pad_val], [pad_val, pad_val], [.2, .4], [.1, .4]]))
  np.testing.assert_allclose(
    out["b"],
    np.asarray([[pad_val, pad_val], [pad_val, pad_val], [.1, .2], [.6, .6]]))
  out = buffer.update(
    signals={"a": [[.1, .1,], [.1, .1,], [.1, .1]], "c": [[.1], [.1]]}, dt=2)
  assert len(out) == 1
  np.testing.assert_allclose(
    out["a"],
    np.asarray([[.2, .4], [.1, .25], [.1, .1], [.1, .1]]))



# Copyright (c) 2024 Rouast Labs
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

class VitalLensAPIKeyError(Exception):
  """Exception raised for errors related to the API key."""
  def __init__(self, message="A valid API key is required to use Method.VITALLENS. If you signed up recently, please try again in a minute to allow your API key to become active. Otherwise, head to https://www.rouast.com/api to get a free API key."):
    self.message = message
    super().__init__(self.message)

class VitalLensAPIQuotaExceededError(Exception):
  """Exception raised if quota exceeded."""
  def __init__(self, message="The quota or rate limit associated with your API Key may have been exceeded. Check your account at https://www.rouast.com/api and consider changing to a different plan."):
    self.message = message
    super().__init__(self.message)

class VitalLensAPIError(Exception):
  """Exception raised for internal API errors."""
  def __init__(self, message="Bad request or an error occured in the API."):
    self.message = message
    super().__init__(self.message)



# Copyright (c) 2024 Rouast Labs
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import base64
import concurrent.futures
import math
import numpy as np
from prpy.numpy.core import standardize
from prpy.numpy.face import get_roi_from_det
from prpy.numpy.filters import detrend, moving_average
from prpy.numpy.image import probe_image_inputs, parse_image_inputs
from prpy.numpy.interp import interpolate_filtered
from prpy.numpy.physio import detrend_lambda_for_hr_response, detrend_lambda_for_rr_response
from prpy.numpy.physio import moving_average_size_for_hr_response, moving_average_size_for_rr_response
from prpy.numpy.utils import enough_memory_for_ndarray
import json
import logging
import requests
from typing import Union, Tuple

from vitallens.constants import API_MAX_FRAMES, API_URL, API_OVERLAP
from vitallens.constants import CALC_HR_MIN_WINDOW_SIZE, CALC_RR_MIN_WINDOW_SIZE
from vitallens.enums import Mode
from vitallens.errors import VitalLensAPIKeyError, VitalLensAPIQuotaExceededError, VitalLensAPIError
from vitallens.methods.rppg_method import RPPGMethod
from vitallens.signal import reassemble_from_windows, assemble_results
from vitallens.utils import check_faces_in_roi

class VitalLensRPPGMethod(RPPGMethod):
  """RPPG method using the VitalLens API for inference"""
  def __init__(
      self,
      config: dict,
      mode: Mode,
      api_key: str
    ):
    """Initialize the `VitalLensRPPGMethod`
    
    Args:
      config: The configuration dict
      mode: The operation mode
      api_key: The API key
    """
    super(VitalLensRPPGMethod, self).__init__(config=config, mode=mode)
    self.api_key = api_key
    self.model = config['model']
    self.input_size = config['input_size']
    self.n_inputs = config['n_inputs']
    self.roi_method = config['roi_method']
    self.signals = config['signals']
    if mode == Mode.BURST:
      self.state = None
      self.input_buffer = None
  def __call__(
      self,
      inputs: Union[np.ndarray, str],
      faces: np.ndarray,
      fps: float = None,
      override_fps_target: float = None,
      override_global_parse: bool = None
    ) -> Tuple[dict, dict, dict, dict, np.ndarray]:
    """Estimate vitals from video frames using the VitalLens API.

    Args:
      inputs: The video to analyze. Either a np.ndarray of shape (n_frames, h, w, 3)
        in unscaled uint8 RGB format, or a path to a video file.
      faces: The face detection boxes as np.int64. Shape (n_frames, 4) in form (x0, y0, x1, y1)
      fps: The rate at which video was sampled.
      override_fps_target: Override the method's default inference fps (optional).
      override_global_parse: If True, always use global parse. If False, don't use global parse.
        If None, choose based on video.
    Returns:
      Tuple of
       - out_data: The estimated data/value for each signal.
       - out_unit: The estimation unit for each signal.
       - out_conf: The estimation confidence for each signal.
       - out_note: An explanatory note for each signal.
       - live: The face live confidence. Shape (n_frames,)
    """
    inputs_shape, fps, video_issues = probe_image_inputs(inputs, fps=fps)
    # Check the number of frames to be processed
    inputs_n = inputs_shape[0]
    fps_target = override_fps_target if override_fps_target is not None else self.fps_target
    expected_ds_factor = max(round(fps / fps_target), 1)
    expected_ds_n = math.ceil(inputs_n / expected_ds_factor)
    # Check if we should parse the video globally
    video_fits_in_memory = enough_memory_for_ndarray(
      shape=(expected_ds_n, self.input_size, self.input_size, 3), dtype=np.uint8,
      max_fraction_of_available_memory_to_use=0.1)
    global_face = faces[np.argmin(np.linalg.norm(faces - np.median(faces, axis=0), axis=1))]
    global_roi = get_roi_from_det(
      global_face, roi_method=self.roi_method, clip_dims=(inputs_shape[2], inputs_shape[1]))
    global_faces_in_roi = check_faces_in_roi(faces=faces, roi=global_roi, percentage_required_inside_roi=(0.6, 1.0))
    global_parse = isinstance(inputs, str) and video_fits_in_memory and (video_issues or global_faces_in_roi)
    if override_global_parse is not None: global_parse = override_global_parse
    if global_parse:
      # Parse entire video for inference globally
      frames, _, _, _, idxs = parse_image_inputs(
        inputs=inputs, fps=fps, roi=global_roi, target_size=self.input_size, target_fps=fps_target,
        preserve_aspect_ratio=False, library='prpy', scale_algorithm='bilinear', 
        trim=None, allow_image=False, videodims=True)
    else:
      frames = inputs
    # Longer videos are split up with small overlaps
    n_splits = 1 if expected_ds_n <= API_MAX_FRAMES else math.ceil((expected_ds_n - API_MAX_FRAMES) / (API_MAX_FRAMES - API_OVERLAP)) + 1
    split_len = inputs_n if n_splits == 1 else math.ceil((inputs_n + (n_splits-1) * API_OVERLAP) / n_splits)
    start_idxs = [i * (split_len - API_OVERLAP) for i in range(n_splits)]
    end_idxs = [min(start + split_len, inputs_n) for start in start_idxs]
    start_idxs = [max(0, end - split_len) for end in end_idxs]
    logging.info(
      f"Analyzing video of {expected_ds_n} frames using {n_splits} request{'s' if n_splits > 1 else ''}. "
      f"Using {n_splits * split_len} frames in total{' including overlaps' if n_splits > 1 else ''}...")
    # Process the splits in parallel
    with concurrent.futures.ThreadPoolExecutor() as executor:
      results = list(executor.map(lambda i: self.process_api_batch(
        batch=i, n_batches=n_splits, inputs=frames, inputs_shape=inputs_shape,
        faces=faces, fps_target=fps_target, fps=fps, global_parse=global_parse,
        start=None if n_splits == 1 else start_idxs[i],
        end=None if n_splits == 1 else end_idxs[i]), range(n_splits)))
    # Aggregate the results
    sig_results, conf_results, live_results, idxs_results = zip(*results)
    sig_ds, idxs = reassemble_from_windows(x=sig_results, idxs=idxs_results)
    conf_ds, _ = reassemble_from_windows(x=conf_results, idxs=idxs_results)
    live_ds = reassemble_from_windows(x=np.asarray(live_results)[:,np.newaxis], idxs=idxs_results)[0][0]
    # Interpolate to original sampling rate
    sig = interpolate_filtered(t_in=idxs, s_in=sig_ds, t_out=np.arange(inputs_n), axis=1, extrapolate=True)
    conf = interpolate_filtered(t_in=idxs, s_in=conf_ds, t_out=np.arange(inputs_n), axis=1, extrapolate=True)
    live = interpolate_filtered(t_in=idxs, s_in=live_ds, t_out=np.arange(inputs_n), axis=0, extrapolate=True)
    # Filter only in batch mode (2, n_frames)
    if self.op_mode == Mode.BATCH:
      sig = np.asarray([self.postprocess(p, fps, type=name) for p, name in zip(sig, ['ppg', 'resp'])])
    # Assemble and return the results
    return assemble_results(sig=sig,
                            conf=conf,
                            live=live,
                            fps=fps,
                            train_sig_names=['ppg_waveform', 'respiratory_waveform'],
                            pred_signals=self.signals,
                            method_name=self.model,
                            min_t_hr=CALC_HR_MIN_WINDOW_SIZE,
                            min_t_rr=CALC_RR_MIN_WINDOW_SIZE,
                            can_provide_confidence=True)
  def process_api_batch(
      self,
      batch: int,
      n_batches: int,
      inputs: Tuple[np.ndarray, str],
      inputs_shape: tuple,
      faces: np.ndarray,
      fps_target: float,
      start: int = None,
      end: int = None,
      fps: float = None,
      global_parse: bool = False
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Process a batch of frames with the VitalLens API.

    Args:
      batch: The number of this batch.
      n_batches: The total number of batches.
      inputs: The video to analyze. Either a np.ndarray of shape (n_frames, h, w, 3)
        with a sequence of frames in unscaled uint8 RGB format, or a path to a video file.
      inputs_shape: The original shape of the inputs.
      faces: The face detection boxes as np.int64. Shape (n_frames, 4) in form (x0, y0, x1, y1)
      fps_target: The target frame rate at which to run inference.
      start: The index of first frame of the video to analyze in this batch.
      end: The index of the last frame of the video to analyze in this batch.
      fps: The frame rate of the input video. Required if type(video) == np.ndarray
      global_parse: Flag that indicates whether video has already been parsed.
    Returns:
      Tuple of
       - sig: Estimated signals. Shape (n_sig, n_frames)
       - conf: Estimation confidences. Shape (n_sig, n_frames)
       - live: Liveness estimation. Shape (n_frames,)
       - idxs: Indices in inputs that were processed. Shape (n_frames)
    """
    logging.debug(f"Batch {batch}/{n_batches}...")
    # Trim face detections to batch if necessary
    if start is not None and end is not None:
      faces = faces[start:end]
    # Choose representative face detection
    face = faces[np.argmin(np.linalg.norm(faces - np.median(faces, axis=0), axis=1))]
    roi = get_roi_from_det(
      face, roi_method=self.roi_method, clip_dims=(inputs_shape[2], inputs_shape[1]))
    if not check_faces_in_roi(faces=faces, roi=roi):
      logging.warning("Large face movement detected")
    if global_parse:
      # Inputs have already been parsed globally.
      assert isinstance(inputs, np.ndarray)
      frames_ds = inputs
      ds_factor = math.ceil(inputs_shape[0] / frames_ds.shape[0])
      # Trim frames to batch if necessary
      if start is not None and end is not None:
        start_ds = start // ds_factor
        end_ds = math.ceil((end-start)/ds_factor) + start_ds
        frames_ds = frames_ds[start_ds:end_ds]
        idxs = list(range(start, end, ds_factor))
      else:
        idxs = list(range(0, inputs_shape[0], ds_factor))
    else:
      # Buffer inputs for burst mode
      if self.op_mode == Mode.BURST:
        # Inputs in burst mode are always np.ndarray
        if self.state is not None:
          # State has been initialized
          assert self.input_buffer is not None
          if inputs.shape[1:] != self.input_buffer.shape[1:]:
            raise ValueError("In burst mode, input dimensions must be consistent.")
          inputs = np.concatenate([self.input_buffer, inputs], axis=0)
        self.input_buffer = inputs[-(self.n_inputs-1):]
      # Inputs have not been parsed globally. Parse the inputs
      frames_ds, _, _, ds_factor, idxs = parse_image_inputs(
        inputs=inputs, fps=fps, roi=roi, target_size=self.input_size, target_fps=fps_target,
        preserve_aspect_ratio=False, library='prpy', scale_algorithm='bilinear', 
        trim=(start, end) if start is not None and end is not None else None,
        allow_image=False, videodims=True)
    # Make sure we have the correct number of frames
    idxs = np.asarray(idxs)
    expected_n = math.ceil(((end-start) if start is not None and end is not None else inputs_shape[0]) / ds_factor)
    if (self.op_mode == Mode.BURST and self.state is not None): expected_n += (self.n_inputs - 1)
    if frames_ds.shape[0] != expected_n or idxs.shape[0] != expected_n:
      raise ValueError("Unexpected number of frames returned. Try to set `override_global_parse` to `True` or `False`.")
    # Prepare API header and payload
    # -- by not sending fps information, ask endpoint not to do any processing
    headers = {"x-api-key": self.api_key}
    payload = {"video": base64.b64encode(frames_ds.tobytes()).decode('utf-8'), "origin": "vitallens-python"}
    if self.op_mode == Mode.BURST:
      if self.state is not None:
        # State and frame buffer have been initialized
        assert self.input_buffer is not None
        payload["state"] = base64.b64encode(self.state.astype(np.float32).tobytes()).decode('utf-8')
        # Adjust idxs
        idxs = idxs[(self.n_inputs-1):] - (self.n_inputs-1)
        logging.debug(f"Providing state, which means that {self.n_inputs-1} less frames will be used and results for {self.n_inputs-1} less frames will be returned.")
    # Ask API to process video
    response = requests.post(API_URL, headers=headers, json=payload)
    response_body = json.loads(response.text)
    # Check if call was successful
    if response.status_code != 200:
      logging.error(f"Error {response.status_code}: {response_body['message']}")
      if response.status_code == 403:
        raise VitalLensAPIKeyError()
      elif response.status_code == 429:
        raise VitalLensAPIQuotaExceededError()
      elif response.status_code == 400:
        raise VitalLensAPIError(f"Parameters missing: {response_body['message']}")
      elif response.status_code == 422:
        raise VitalLensAPIError(f"Issue with provided parameters: {response_body['message']}")
      elif response.status_code == 500:
        raise VitalLensAPIError(f"Error occurred in the API: {response_body['message']}")
      else:
        raise Exception(f"Error {response.status_code}: {response_body['message']}")
    # Parse response
    sig_ds = np.stack([
      np.asarray(response_body["vital_signs"]["ppg_waveform"]["data"]),
      np.asarray(response_body["vital_signs"]["respiratory_waveform"]["data"]),
    ], axis=0)
    conf_ds = np.stack([
      np.asarray(response_body["vital_signs"]["ppg_waveform"]["confidence"]),
      np.asarray(response_body["vital_signs"]["respiratory_waveform"]["confidence"]),
    ], axis=0)
    live_ds = np.asarray(response_body["face"]["confidence"])
    if self.op_mode == Mode.BURST:
      self.state = np.asarray(response_body["state"]["data"], dtype=np.float32)
    return sig_ds, conf_ds, live_ds, idxs
  def postprocess(
      self,
      sig: np.ndarray,
      fps: float,
      type: str = 'ppg',
      filter: bool = True
    ) -> np.ndarray:
    """Apply filters to the estimated signal. 
    Args:
      sig: The estimated signal. Shape (n_frames,)
      fps: The rate at which video was sampled. Scalar
      type: The signal type - either 'pulse' or 'resp'.
      filter: Whether to apply filters
    Returns:
      x: The filtered signal. Shape (n_frames,)
    """
    n_frames = sig.shape[-1]
    if filter:
      # Get filter parameters
      if type == 'ppg':
        size = moving_average_size_for_hr_response(fps)
        Lambda = detrend_lambda_for_hr_response(fps)
      elif type == 'resp':
        size = moving_average_size_for_rr_response(fps)
        Lambda = detrend_lambda_for_rr_response(fps)
      else:
        raise ValueError(f"Type {type} not implemented!")
      if sig.shape[-1] < 4 * API_MAX_FRAMES:
        # Detrend only for shorter videos for performance reasons
        sig = detrend(sig, Lambda)
      # Moving average
      sig = moving_average(sig, size)
      # Standardize
      sig = standardize(sig)
    # Return
    assert sig.shape == (n_frames,)
    return sig
  def reset(self):
    """Reset"""
    if self.op_mode == Mode.BURST:
      self.state = None
      self.input_buffer = None



# Copyright (c) 2024 Rouast Labs
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from dotenv import load_dotenv
import os
load_dotenv()

# Vitals estimation window sizes [s]
CALC_HR_MIN_WINDOW_SIZE = 2
CALC_HR_WINDOW_SIZE = 10
CALC_RR_MIN_WINDOW_SIZE = 4
CALC_RR_WINDOW_SIZE = 30

# API settings
API_MIN_FRAMES = 16
API_MAX_FRAMES = 900
API_OVERLAP = 30
API_URL = "https://api.rouast.com/vitallens-v3/file"
if 'API_URL' in os.environ:
  API_URL = os.getenv('API_URL')

# Video error message
VIDEO_PARSE_ERROR = "Unable to parse input video. There may be an issue with the video file."

# Disclaimer message
DISCLAIMER = "The provided values are estimates and should be interpreted according to the provided confidence levels ranging from 0 to 1. The VitalLens API is not a medical device and its estimates are not intended for any medical purposes."
